import rasterio as rio
from rasterio import warp
import numpy as np

def _wrap_x_coord(xcoords):
    '''
    Wraps longitude coords beyond -180 / 180

    Parameters
    -----------
    xcoords: ndarray or value
        coordinate or coordinates to wrap.

    Returns
    --------
    wrapped: ndarray or value
        wrapped coordinate(s)
    '''
    return ((xcoords + 180) % 360) - 180

def winding_order(boundsArr):
    '''
    returns False if CCW; True is CW
    (ie, it crosses the dateline and is inverted when unprojected)
    '''
    shiftBounds = np.roll(boundsArr, -1, 0)
    delta = (shiftBounds[:, 0] - boundsArr[:, 0]) * (shiftBounds[:, 1] + boundsArr[:, 1])

    return np.sum(delta) > 0


def densify(boundsArr, dens=4):
    '''
    Adds points between the bounds corners
    '''
    denseBounds = np.zeros(
        ((boundsArr.shape[0]) * dens, 2),
        dtype=boundsArr.dtype)

    denseBounds[::dens] += boundsArr
    diffs = (boundsArr - np.roll(boundsArr, -1, 0)) / float(dens)
    for i in range(1, dens):
        denseBounds[i::dens] += boundsArr - (diffs * i)

    return denseBounds


def make_bounds_array(bounds):
    '''
    Array of raster bounds
    in counter-clockwise
    fashion
    '''
    return np.array([
        [bounds.left, bounds.bottom],
        [bounds.right, bounds.bottom],
        [bounds.right, bounds.top],
        [bounds.left, bounds.top]
        ])


def transform_bounds(boundsArr, crs):
    if not crs:
        raise ValueError('Input raster must have a CRS')

    if 'init' in crs and crs['init'] == 'epsg:4326':
        boundsArr[:, 0] = _wrap_x_coord(boundsArr[:, 0])
    else:
        boundsArr = np.dstack(warp.transform(
                crs,
                {'init': 'epsg:4326'},
                boundsArr[:, 0],
                boundsArr[:, 1]))[0]

    return boundsArr


def crosses_dateline(fpath):
    '''
    Checks if a raster crosses the dateline after unprojection to epsg 4326
    '''
    with rio.open(fpath) as src:
        denseBounds = densify(make_bounds_array(src.bounds))
        unprojectedBounds = transform_bounds(denseBounds, src.crs)

        return winding_order(unprojectedBounds)



#!/usr/bin/env python
import rasterio

def tiled(sources, equal_blocks=True):
    """ Tests if raster sources have the same tiling
    optionally assert that their block sizes are equal (default: True)
    """
    blocksize = None
    with rasterio.Env():
        for source in sources:
            with rasterio.open(source) as src:

                if not src.profile['tiled']:
                    return (False, "Source(s) are not internally tiled")

                if equal_blocks:
                    this_blocksize = (src.profile['blockxsize'],
                                      src.profile['blockysize'])
                    if blocksize:
                        if blocksize != this_blocksize:
                            return (False, "Blocksizes are not equal")
                    else:
                        blocksize = this_blocksize

    return (True, "Tiling checks out")


def aligned(sources):
    """ Tests if raster sources are aligned
    i.e. have the same transform and shape
    (and by extension the same resolution and extent)
    """
    atransform = None
    shape = None
    with rasterio.Env():
        for source in sources:
            with rasterio.open(source) as src:

                if atransform:
                    if src.affine != atransform:
                        return (False, "Affine transforms are not aligned")
                else:
                    atransform = src.affine

                if shape:
                    if src.shape != shape:
                        return (False, "Sources have different shape")
                else:
                    shape = src.shape

    return (True, "Sources have the same shape and transform")



#!/usr/bin/env python

from affine import Affine
import click
import numpy as np
import rasterio
from rasterio.warp import reproject
try:
    from rasterio.warp import RESAMPLING
except ImportError:
    from rasterio.enums import Resampling as RESAMPLING
from rasterio.coords import BoundingBox

from .utils import exception_raiser


def affaux(up):
    return Affine(1, 0, 0, 0, -1, 0), Affine(up, 0, 0, 0, -up, 0)


def upsample_array(bidx, up, fr, to):
    upBidx = np.empty(
        (bidx.shape[0] * up, bidx.shape[1] * up), dtype=bidx.dtype)

    reproject(
        bidx, upBidx,
        src_transform=fr,
        dst_transform=to,
        src_crs="EPSG:3857",
        dst_crs="EPSG:3857",
        resampling=RESAMPLING.bilinear)

    return upBidx

def array_compare(arr1, arr2, valueFilter=0, countFilter=0, debug=False):
    diffArr = np.abs(arr1.astype(np.int64) -
                     arr2.astype(np.int64)).astype(arr1.dtype)

    diffSpots = np.where(diffArr > valueFilter)

    diffCount = diffSpots[0].size

    if debug and diffCount > countFilter:
        rows, cols = arr1.shape
        divver = int(np.mean([rows, cols])) / 60 + 1
        mapa = ['-', 'X']
        diffPlot = (np.histogram2d(diffSpots[0], diffSpots[1], (np.arange(
            0, cols, divver * 2), np.arange(0, rows, divver)))[0] > 0).astype(np.uint8)
        click.secho('\n'.join([''.join([mapa[i] for i in row])
                               for row in diffPlot]), fg='red')

    return diffCount, diffCount > countFilter


def make_fill_array(height, width, downsample, dtype):
    return np.zeros(
        (int(height / downsample), int(width / downsample)),
        dtype
    )


def compare_properties(src1, src2, properties, tol=1e-6):
    noMatch = []
    for prop in properties:
        a = getattr(src1, prop)
        b = getattr(src2, prop)
        equal = True
        if isinstance(a, BoundingBox) and isinstance(b, BoundingBox):
            for coord_a, coord_b in zip(tuple(a), tuple(b)):
                if abs(coord_a - coord_b) > tol:
                    equal = False
                    break
        elif isinstance(a, float) and isinstance(b, float):
            if abs(a - b) > tol:
                equal = False
        else:
            if a != b:
                equal = False

        if not equal:
            noMatch.append({
                prop: {
                    'src1': a,
                    'src2': b
                }
            })

    if not len(noMatch):
        noMatch = None

    return noMatch


def compare(srcpath1, srcpath2, max_px_diff=0, upsample=1, downsample=1,
            compare_masked=True, no_stderr=False, debug=False, flex_mode=False):
    with rasterio.Env(GTIFF_IMPLICIT_JPEG_OVR=False):
        with rasterio.open(srcpath1) as src1:
            with rasterio.open(srcpath2) as src2:

                count1 = src1.count
                count2 = src2.count
                compareAlpha = 1

                if flex_mode and [count1, count2].count(3) != 0:
                    props = ['crs', 'driver', 'bounds', 'height', 'width', 'shape']
                    if src1.count * src2.count != 12 or src1.count + src2.count != 7:
                        exception_raiser(
                            "In flex mode, %s and %s must 3 and 4, or 4 and 3 bands "
                            "respectively (received %s and %s)" % (
                                srcpath1, srcpath2, src1.count, src2.count), no_stderr)
                else:
                    props = ['count', 'crs', 'dtypes', 'driver', 'bounds',
                            'height', 'width', 'shape', 'nodatavals']
                propCompare = compare_properties(src1, src2, props)

                if propCompare:
                    exception_raiser(propCompare, no_stderr)

                if compare_masked and src1.count == 4 and not flex_mode:
                    # create arrays for decimated reading
                    masked_1 = make_fill_array(
                        src1.height, src1.width, downsample, src1.meta['dtype'])
                    masked_2 = make_fill_array(
                        src2.height, src2.width, downsample, src2.meta['dtype'])

                    src1.read(4, out=masked_1, masked=False)
                    src2.read(4, out=masked_2, masked=False)
                    compareAlpha = 0
                    difference, aboveThreshold = array_compare(
                        masked_1, masked_2, 16, max_px_diff, debug)

                    if aboveThreshold:
                        exception_raiser(
                            'Mask has %s pixels that vary by more than 16' % (difference), no_stderr)

                elif compare_masked and flex_mode:
                    masked_1 = make_fill_array(
                        src1.height, src1.width, downsample, src1.meta['dtype'])
                    masked_2 = make_fill_array(
                        src2.height, src2.width, downsample, src2.meta['dtype'])

                    src1.read_masks(1, out=masked_1)
                    src2.read_masks(1, out=masked_2)
                    compareAlpha = 0

                    difference, aboveThreshold = array_compare(
                        masked_1, masked_2, 16, max_px_diff, debug)

                    if aboveThreshold:
                        exception_raiser(
                            'Mask has %s pixels that vary by more than 16' % (difference),
                            no_stderr)

                for bidx in range(1, count1 + compareAlpha):
                    # create arrays for decimated reading
                    band1 = make_fill_array(
                        src1.height, src1.width, downsample, src1.meta['dtype'])
                    band2 = make_fill_array(
                        src2.height, src2.width, downsample, src2.meta['dtype'])

                    src1.read(bidx, out=band1, masked=False)
                    band1 = band1.astype(np.int16)

                    src2.read(bidx, out=band2, masked=False)
                    band2 = band2.astype(np.int16)

                    if compare_masked and src1.count == 4:
                        band1[masked_1 != 255] = 0
                        band2[masked_1 != 255] = 0

                    if upsample > 1:
                        toAff, frAff = affaux(upsample)
                        band1 = upsample_array(band1, upsample, frAff, toAff)
                        band2 = upsample_array(band2, upsample, frAff, toAff)

                    difference, aboveThreshold = array_compare(
                        band1, band2, 16, max_px_diff, debug)

                    if aboveThreshold:
                        exception_raiser('Band %s has %s pixels that vary by more than 16' % (
                            bidx, difference), no_stderr)

    click.echo("ok - %s is similar to within %s pixels of %s" %
               (srcpath1, max_px_diff, srcpath2))



from .compare import compare, array_compare, compare_properties
from .empty import is_empty
from .aligned import tiled, aligned
from .crosses_dateline import crosses_dateline, winding_order, densify, make_bounds_array, transform_bounds

__all__ = [is_empty, compare, array_compare, compare_properties,
           tiled, aligned, crosses_dateline, winding_order, densify, make_bounds_array, transform_bounds]



import click
import sys

import raster_tester


@click.group()
def cli():
    pass


@click.command("compare")
@click.argument("input_1", type=click.Path(exists=True))
@click.argument("input_2", type=click.Path(exists=True))
@click.option("--pixel-threshold", "-p", type=int, default=0,
              help='Threshold for pixel diffs [default=0]')
@click.option("--downsample", "-d", type=int, default=1,
              help='Downsample via decimated read for faster comparison, and to handle variation in compression artifacts [default=1]')
@click.option("--upsample", "-u", type=int, default=1,
              help='Upsample to handle variation in compression artifacts [default=1]')
@click.option("--compare-masked", is_flag=True,
              help='Only compare masks + unmasked areas of RGBA rasters')
@click.option("--no-error", is_flag=True,
              help='Compare in non stderr mode: echos "(ok|not ok) - <input_1> is (within|not within) <pixel-threshold> pixels of <input 2>"')
@click.option("--debug", is_flag=True,
              help='Print ascii preview of errors')
@click.option("--flex-mode", is_flag=True,
              help='Allow comparison of masked RGB + RGBA')
def compare(input_1, input_2, pixel_threshold, upsample, downsample,
            no_error, compare_masked, debug, flex_mode):
    raster_tester.compare(
        input_1, input_2, pixel_threshold, upsample, downsample,
        compare_masked, no_error, debug, flex_mode)


cli.add_command(compare)


@click.command("isempty")
@click.argument("input_1", type=click.Path(exists=True))
@click.option('--bidx', '-b', default=4,
              help="Bands to blob [default = 4]")
@click.option("--randomize", is_flag=True,
              help='iterate through windows in a psuedorandom fashion')
def isempty(input_1, randomize, bidx):
    empty = raster_tester.is_empty(input_1, randomize, bidx, )
    exits = {
        True: ("is empty", 0),
        False: ("is not empty", 1)
    }

    message, eCode = exits[empty]

    click.echo("%s %s" % (input_1, message))
    sys.exit(eCode)

cli.add_command(isempty)


@click.command("isaligned")
@click.argument('sources', required=True, nargs=-1)
def isaligned(sources):
    aligned, msg = raster_tester.aligned(sources)
    if aligned:
        click.echo("ok: {} are aligned ({})".format(', '.join(sources), msg))
        sys.exit(0)
    else:
        click.echo("not ok: {} are not aligned ({})".format(', '.join(sources), msg))
        sys.exit(1)

cli.add_command(isaligned)


@click.command("istiled")
@click.argument('sources', required=True, nargs=-1)
@click.option('--blocksize/--no-blocksize', is_flag=True, default=True,
              help="assert that sources are internally tiled")
def istiled(sources, blocksize):
    result, msg = raster_tester.tiled(sources, blocksize)
    if result:
        click.echo("ok: {} are tiled ({})".format(', '.join(sources), msg))
        sys.exit(0)
    else:
        click.echo("not ok: {} are not all tiled ({})".format(', '.join(sources), msg))
        sys.exit(1)

cli.add_command(istiled)

@click.command('crossesdateline')
@click.argument('input', type=click.Path(exists=True))
def crossesdateline(input):
    result = raster_tester.crosses_dateline(input)
    if result:
        click.echo('%s crosses dateline; exit 1' % (input))
        sys.exit(1)
    else:
        click.echo('%s does not cross dateline; exit 0' % (input))
        sys.exit(0)

cli.add_command(crossesdateline)

if __name__ == "__main__":
    cli()



#!/usr/bin/env python
import sys

import click

def exception_raiser(message, no_stderr):
    if no_stderr:
        click.echo("not ok - %s" % (message))
        sys.exit(0)
    else:
        raise ValueError(message)



import click, sys

import numpy as np
import rasterio as rio

def array_hasdata(arr):
    return np.any(arr)

def is_empty(input_path, randomize, bidx):
    with rio.open(input_path) as src:
        windows = [window for ij, window in src.block_windows()]

        if randomize:
            np.random.shuffle(windows)

        empty = True

        for window in windows:
            if array_hasdata(src.read(bidx, window=window)):
                empty = False
                break

        return empty
